#!/usr/bin/python
# -*- coding: utf-8 -*-
#This file is part of pydsl.
#
#pydsl is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydsl is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydsl.  If not, see <http://www.gnu.org/licenses/>.


""" Directory storage """

from .Memory import Memory
from pydsl.Memory.File.Python import getFileTuple
from pydsl.Abstract import InmutableDict
import logging
LOG = logging.getLogger(__name__)

def _isGDLFileName(path):
    return path.endswith(".bnf")

def _isRELFileName(path):
    return path.endswith(".re")

def _isBoardFileName(path):
    return path.endswith(".board")

class DirStorage(Memory):
    """A collection of elements stored inside a directory"""

    def __init__(self, dirpath, allowedextensions=(".py", ".bnf", ".re", ".board")):
        Memory.__init__(self)
        self.path = dirpath
        self._allowedextensions = allowedextensions
        from pydsl.Memory.Search.Searcher import MemorySearcher

        self._searcher = MemorySearcher(self)

    def __iter__(self):
        self.index = 0
        self.cache = []
        for filename in self.all_files():
            try:
                self.cache.append(self.summary_from_filename(filename))
            except (AttributeError,ImportError) as e:
                LOG.debug("Error while loading %s file summary" % filename )
        return self

    def next(self):
        try:
            result = self.cache[self.index]
        except IndexError:
            raise StopIteration
        self.index += 1
        return result
        

    def summary_from_filename(self, modulepath):
        (_, _, fileBaseName, ext) = getFileTuple(modulepath)
        if _isRELFileName(modulepath):
            result =  {"iclass":"re","identifier":fileBaseName, "filepath":modulepath}
        elif _isGDLFileName(modulepath):
            result = {"iclass":"BNFGrammar","identifier":fileBaseName, "filepath":modulepath}
        elif _isBoardFileName(modulepath):
            result = {"iclass":"Board", "identifier":fileBaseName, "filepath":modulepath}
        else:
            from pydsl.Memory.File.Python import summary_python_file
            result = summary_python_file(modulepath)
        return InmutableDict(result)

    def all_files(self):
        import glob
        if self._allowedextensions:
            for extension in self._allowedextensions:
                searchstring = self.path + "*" + extension
                tmpresult = glob.glob(searchstring)
                for result in tmpresult:
                    if result.endswith("__init__.py"):
                        continue
                    yield result 
        else:
            searchstring = self.path + "*" 
            for result in glob.glob(searchstring):
                yield result 


    def all_names(self):
        """Generates all Static Ids"""
        for fullname in self.all_files():
            (_, _, fileBaseName, fileExtension) = getFileTuple(fullname)
            if self._allowedextensions and fileExtension not in self._allowedextensions:
                continue
            yield fileBaseName.split(".")[0]

    def _load_module_from_library(self, identifier):
        try:
            import imp
            moduleobject = imp.load_source(identifier, self.path + "/" + identifier + ".py")
        except (ImportError, IOError):
            pass
        else:
            return moduleobject
        raise ImportError

    def load(self, name, **kwargs):
        resultlist = self._searcher.search(name)
        if len(resultlist) > 1:
            LOG.error("Found two or more matches, FIXME: processing the first, should raise exception")
        if len(resultlist) == 0:
            raise KeyError(self.__class__.__name__ + name)
        filepath = list(resultlist)[0]["filepath"]
        if _isRELFileName(filepath):
            from pydsl.Memory.File.Regexp import load_re_from_file
            return load_re_from_file(filepath)
        if _isGDLFileName(filepath):
            from pydsl.Memory.File.BNF import load_bnf_file
            return load_bnf_file(filepath)
        if _isBoardFileName(filepath):
            from pydsl.Memory.File.Board import load_board_file
            return load_board_file(filepath)
        from pydsl.Memory.File.Python import load_python_file
        return load_python_file(filepath, **kwargs)

    def __contains__(self, key):
        return key in self.all_names()

    def provided_iclasses(self):
        return set([x['iclass'] for x in self])




#!/usr/bin/python
# -*- coding: utf-8 -*-
#This file is part of pydsl.
#
#pydsl is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydsl is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydsl.  If not, see <http://www.gnu.org/licenses/>.

"""Base Lexer classes"""

__author__ = "Nestor Arocha"
__copyright__ = "Copyright 2008-2013, Nestor Arocha"
__email__ = "nesaro@gmail.com"

import logging
LOG = logging.getLogger(__name__)
from pydsl.Memory.Loader import load_checker
finalchar = "EOF"
unknownchar = "UNKNOWN"


class Lexer(object):
    """Lexer follows an alphabet definition, which is like a grammar definition but generates a list of tokens and it is always Readable using a regular grammar"""
    def __init__(self, generate_unknown=False):
        self.load(None)
        self.generate_unknown = generate_unknown

    @property
    def current(self):
        """Returns the element under the cursor"""
        try:
            return self.string[self.index]
        except IndexError:
            return finalchar

    def load(self, string):
        self.string = string
        self.index = 0

    def consume(self):
        self.index += 1

    def match(self, char):
        if self.current == char:
            self.consume()
        else:
            raise Exception

    def nextToken(self):
        raise NotImplementedError

    def __call__(self, string):# -> "TokenList":
        """Tokenizes input, generating a list of tokens"""
        self.string = string
        result = []
        while True:
            nt = self.nextToken()
            result.append(nt)
            if nt[0] == "EOF_TYPE":
                return result


class BNFLexer(Lexer):
    """Generates a Lexer from a BNFGrammar instance"""
    def __init__(self, bnfgrammar):
        Lexer.__init__(self)
        self.symbollist = bnfgrammar.getTerminalSymbols()

    @property
    def current(self):
        """Returns the element under the cursor until the end of the string"""
        try:
            return self.string[self.index:]
        except IndexError:
            return finalchar

    def nextToken(self):
        while self.current != finalchar:
            validelements = [x for x in self.symbollist if self.current[0] in x.first()]
            if not validelements:
                if not self.generate_unknown:
                    raise Exception("Not found")
                string = self.current[0]
                self.consume()
                return unknownchar, string
            elif len(validelements) == 1:
                element = validelements[0]
                string = self.current[:len(element)]
                for _ in range(len(element)):
                    self.consume()
                return validelements[0].name, string
            else:
                raise Exception("Multiple choices")

        return "EOF_TYPE", ""

class AlphabetDictLexer(Lexer):
    def __init__(self, alphabet):
        Lexer.__init__(self)
        self.alphabet = alphabet

    @property
    def current(self):
        """Returns the element under the cursor until the end of the string"""
        try:
            return self.string[self.index:]
        except IndexError:
            return finalchar

    def nextToken(self):
        while self.current:
            validelements = [(x,y) for x,y in self.alphabet.grammardict.items() if self.current[0] in y.first]
            if not validelements:
                if not self.generate_unknown:
                    raise Exception("Not found")
                string = self.current[0]
                self.consume()
                return unknownchar, string
            elif len(validelements) == 1:
                element = validelements[0][1]
                checker = load_checker(element)
                for size in range(element.maxsize or len(self.current), element.minsize, -1):
                    if checker.check(self.current[:size]):
                        break
                else:
                    raise Exception("Nothing consumed")
                string = self.current[:size]
                for _ in range(size):
                    self.consume()
                return validelements[0][0], string
            else:
                raise Exception("Multiple choices")

        return "EOF_TYPE", ""




#!/usr/bin/python
# -*- coding: utf-8 -*-
#This file is part of pydsl.
#
#pydsl is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydsl is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydsl.  If not, see <http://www.gnu.org/licenses/>.

"""Python Function Grammar"""

__author__ = "Nestor Arocha"
__copyright__ = "Copyright 2008-2013, Nestor Arocha"
__email__ = "nesaro@gmail.com"

from .Tool import GrammarTools
import logging
LOG = logging.getLogger(__name__)

class PythonGrammarTools(GrammarTools):
    def __init__(self, dictionary):
        GrammarTools.__init__(self, dictionary)
        self.dictionary = dictionary
        self._matchFun = None
        self._askprop = dictionary.get('propFun')
        self._enumFun = dictionary.get('enumFun')
        self._alphabetFun = dictionary.get('alphabetFun')

    def get_groups(self, word, propertyname):
        if self._askprop is not None:
            return [self._askprop(word, propertyname)]
        return []

    def enum(self):
        if self._enumFun is not None:
            return set(self._enumFun())
        return set()

    def tokenize(self, data):
        if self._iterFun is not None:
            for x in self._iterFun(data):
                yield x

    @property
    def alphabet(self): # -> set:
        if self._alphabetFun is not None:
            return set(self._alphabetFun())
        return set()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#This file is part of pydsl.
#
#pydsl is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydsl is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydsl.  If not, see <http://www.gnu.org/licenses/>.


"""ANLTr grammar format functions"""

__author__ = "Nestor Arocha"
__copyright__ = "Copyright 2008-2013, Nestor Arocha"
__email__ = "nesaro@gmail.com"

import logging
LOG = logging.getLogger(__name__)
from pydsl.Alphabet.Lexer import Lexer, finalchar
import re

#manual lexer

protectedwords = {"grammar": "GRAMMAR",
        "options": "OPTIONS",
        "language": "LANGUAGE",
        "output": "OUTPUT",
        "fragment": "FRAGMENT",
        "Java": "JAVA",
        "AST": "AST",
        "ASTLabelType": "ASTLABELTYPE"}

chardict = {",": "COMMA",
        ":": "COLON",
        ";": "SEMICOLON",
        "|": "VBAR",
        "{": "LCBRACKET",
        "}": "RCBRACKET",
        "(": "LPAR",
        ")": "RPAR",
        "=": "EQUAL",
        "!": "EMARK",
        "?": "QMARK",
        "^": "HAT",
        "*": "STAR",
        "-": "HYPHEN",
        "_": "UNDERSCORE",
        ">": "GT",
        #"'": "QUOTE",
        "&": "AMP",
        ".": "DOT",
        "\\": "ISLASH",
        "+": "PLUS",
        "$": "DOLLAR"}


class ANLTRGrammarLexer(Lexer):
    def matchchar(self, char):
        if not char in chardict:
            raise Exception
        current = self.current
        self.match(char)
        return chardict[char], current

    def comment(self):
        self.match("/")
        self.match("*")
        while True:
            while self.current != "*":
                self.consume()
            self.consume()
            if self.current == "/":
                self.consume()
                break

    def name(self):
        import re
        string = ""
        while self.current != finalchar and re.match("[a-zA-Z]", self.current):
            string += self.current
            self.consume()
        if string in protectedwords:
            return string, protectedwords[string]
        if string == string.lower():
            return string, "PARSERID"
        else:
            return string, "LEXERID"

    def rawstring(self):
        self.match("'")
        string = ""
        while self.current not in [finalchar, "'"]:
            string += self.current
            self.consume()
            if string[-1] == "\\":
                string += self.current
                self.consume()
        self.match("'")
        return string, "STRING"

    def number(self):
        import re
        string = ""
        while self.current != finalchar and re.match("[0-9]", self.current):
            string += self.current
            self.consume()
        return "NUMBER", string

    def nextToken(self):
        while self.current != finalchar:
            if self.current == "/":
                self.comment()
                continue
            elif self.current in [" ", '\n', '\t', '\i']:
                self.consume()
                continue
            elif self.current in chardict.keys():
                return self.matchchar(self.current)
            elif re.match("[0-9]", self.current):
                return self.number()
            elif self.current == "'":
                return self.rawstring()
            elif re.match("[a-zA-Z]", self.current):
                return self.name()
            else:
                raise Exception("Unknown char '%s'" % self.current)
        return "EOF_TYPE", ""


def load_anltr_from_text(text):
    lexer = ANLTRGrammarLexer()
    return lexer(text)


#manual parser
def load_anltr_file(filepath):
    """Converts an anltr .g file into a BNFGrammar instance"""
    with open(filepath, 'r') as mlfile:
        content = mlfile.read()
        return load_anltr_from_text(content)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#This file is part of pydsl.
#
#pydsl is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#pydsl is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with pydsl.  If not, see <http://www.gnu.org/licenses/>.

"""Regular Expression Grammars"""

__author__ = "Nestor Arocha"
__copyright__ = "Copyright 2008-2013, Nestor Arocha"
__email__ = "nesaro@gmail.com"

#There are at least three ways to define a regular grammar. Regexp, FSM table, production rules. At this time, I'm going to combine several regexp rules, each one with a name.

import logging
from .Tool import GrammarTools
LOG = logging.getLogger(__name__)
CHARACTERS = {"a","b","c","d","e"}
ALLSYMBOLS = CHARACTERS.union({".",","})

#TODO: What if Terminal symbols in current grammar scope is a language-grammar-accepted word? We need to tell what terminal symbols mean

def regexp_alphabet_helper(string, endchar = None): # -> set:
    index = 0
    result = set()
    while index < len(string):
        if string[index] == '\\':
            if string[index+1] == '\\':
                result.add('\\')
            elif string[index+1] == 'w':
                result = result.union(CHARACTERS)
            else:
                result.add(string[index+1])
            index += 2
            continue
        elif string[index] == endchar:
            return result, index
        elif string[index] == '.':
            result = result.union(ALLSYMBOLS)
        elif string[index] == '*':  
            pass
        elif string[index] == '?':  
            pass
        elif string[index] == '<':
            newresult, newindex = regexp_alphabet_helper(string[index+1:], ">")
            index += newindex + 1
        elif string[index] == '(':
            if string[index+1:index+3] == "?P":
                index += 2
            newresult, newindex =regexp_alphabet_helper(string[index+1:], ")")
            result = result.union(newresult)
            index += newindex + 1
        else:
            result.add(string[index])
        index += 1
    return result, index

class RegularExpressionGrammarTools(GrammarTools):
    def __init__(self, regexp, flags = ""):
        if isinstance(regexp, str):
            from pydsl.Grammar.Definition import RegularExpressionDefinition
            regexp = RegularExpressionDefinition(regexp)
        GrammarTools.__init__(self, regexp)

    def get_groups(self, word, groupname):
        """Match against specific rule"""
        data = str(word)
        if not data:
            return []
        groupdict = self.gd.match(data).groupdict()
        if not groupname in groupdict:
            return []
        return [(self.gd.match(data).start(groupname),self.gd.match(data).end(groupname))] #Returns a list

    def groups(self):
        return list(self.gd.groupindex.keys())

    def tokenize(self, information):
        """Uses python str iteration"""
        for x in information:
            yield x

    def alphabet(self):
        #FIXME:It is not working with groups
        return regexp_alphabet_helper(self.gd.regexpstr)[0]
